@@ -1,4 +1,4 @@
-# Copyright (c) 2009, 2020, Oracle and/or its affiliates.
+# Copyright (c) 2009, 2021, Oracle and/or its affiliates.
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License, version 2.0, as
@@ -35,7 +35,6 @@ import weakref
 
 from . import errors
 from .abstracts import MySQLCursorAbstract, NAMED_TUPLE_CACHE
-from .catch23 import PY2
 from .constants import ServerFlag
 
 SQL_COMMENT = r"\/\*.*?\*\/"
@@ -115,14 +114,9 @@ def _bytestr_format_dict(bytestr, value_dict):
         if value is None:
             raise ValueError("Unsupported conversion_type: {0}"
                              "".format(groups["conversion_type"]))
-        return bytes(value) if PY2 else value
+        return value
 
     stmt = RE_PY_MAPPING_PARAM.sub(replace, bytestr)
-    if PY2:
-        try:
-            return stmt.decode("utf-8")
-        except UnicodeDecodeError:
-            pass
     return stmt
 
 
@@ -367,6 +361,14 @@ class MySQLCursor(CursorBase):
         except AttributeError:
             return False
 
+    def _check_executed(self):
+        """Check if the statement has been executed.
+
+        Raises an error if the statement has not been executed.
+        """
+        if self._executed is None:
+            raise errors.InterfaceError(ERR_NO_RESULT_TO_FETCH)
+
     def next(self):
         """Used for iterating over the result set."""
         return self.__next__()
@@ -410,10 +412,7 @@ class MySQLCursor(CursorBase):
                 conv = to_mysql(conv)
                 conv = escape(conv)
                 conv = quote(conv)
-                if PY2:
-                    res[key] = conv
-                else:
-                    res[key.encode()] = conv
+                res[key.encode()] = conv
         except Exception as err:
             raise errors.ProgrammingError(
                 "Failed processing pyformat-parameters; %s" % err)
@@ -559,6 +558,10 @@ class MySQLCursor(CursorBase):
                 if psub.remaining != 0:
                     raise errors.ProgrammingError(
                         "Not all parameters were used in the SQL statement")
+            else:
+                raise errors.ProgrammingError(
+                    f"Could not process parameters: {type(params).__name__}({params}),"
+                    " it must be of type list, tuple or dict")
 
         self._executed = stmt
         if multi:
@@ -667,6 +670,7 @@ class MySQLCursor(CursorBase):
                 return None
             stmt = self._batch_insert(operation, seq_params)
             if stmt is not None:
+                self._executed = stmt
                 return self.execute(stmt)
 
         rowcnt = 0
@@ -765,10 +769,17 @@ class MySQLCursor(CursorBase):
             can_consume_results = self._connection._consume_results
             for result in self._connection.cmd_query_iter(call):
                 self._connection._consume_results = False
-                if self._raw:
-                    tmp = MySQLCursorBufferedRaw(self._connection._get_self())
+                if isinstance(self, (MySQLCursorDict,
+                                     MySQLCursorBufferedDict)):
+                    cursor_class = MySQLCursorBufferedDict
+                elif isinstance(self, (MySQLCursorNamedTuple,
+                                       MySQLCursorBufferedNamedTuple)):
+                    cursor_class = MySQLCursorBufferedNamedTuple
+                elif self._raw:
+                    cursor_class = MySQLCursorBufferedRaw
                 else:
-                    tmp = MySQLCursorBuffered(self._connection._get_self())
+                    cursor_class = MySQLCursorBuffered
+                tmp = cursor_class(self._connection._get_self())
                 tmp._executed = "(a result of {0})".format(call)
                 tmp._handle_result(result)
                 if tmp._warnings is not None:
@@ -779,7 +790,12 @@ class MySQLCursor(CursorBase):
             # pylint: enable=W0212
 
             if argnames:
-                select = "SELECT {0}".format(','.join(argtypes))
+                # Create names aliases to be compatible with namedtuples
+                args = [
+                    "{} AS {}".format(name, alias) for name, alias in
+                    zip(argtypes, [arg.lstrip("@_") for arg in argnames])
+                ]
+                select = "SELECT {}".format(",".join(args))
                 self.execute(select)
                 self._stored_results = results
                 return self.fetchone()
@@ -875,12 +891,11 @@ class MySQLCursor(CursorBase):
 
         Returns a tuple or None.
         """
-        row = self._fetch_row()
-        if row:
-            return row
-        return None
+        self._check_executed()
+        return self._fetch_row()
 
     def fetchmany(self, size=None):
+        self._check_executed()
         res = []
         cnt = (size or self.arraysize)
         while cnt > 0 and self._have_unread_result():
@@ -891,8 +906,10 @@ class MySQLCursor(CursorBase):
         return res
 
     def fetchall(self):
+        self._check_executed()
         if not self._have_unread_result():
-            raise errors.InterfaceError("No result set to fetch from.")
+            return []
+
         (rows, eof) = self._connection.get_rows()
         if self._nextrow[0]:
             rows.insert(0, self._nextrow[0])
@@ -995,20 +1012,19 @@ class MySQLCursorBuffered(MySQLCursor):
 
         Returns a tuple or None.
         """
-        row = self._fetch_row()
-        if row:
-            return row
-        return None
+        self._check_executed()
+        return self._fetch_row()
 
     def fetchall(self):
-        if self._rows is None:
-            raise errors.InterfaceError("No result set to fetch from.")
+        if self._executed is None or self._rows is None:
+            raise errors.InterfaceError(ERR_NO_RESULT_TO_FETCH)
         res = []
         res = self._rows[self._next_row:]
         self._next_row = len(self._rows)
         return res
 
     def fetchmany(self, size=None):
+        self._check_executed()
         res = []
         cnt = (size or self.arraysize)
         while cnt > 0:
@@ -1032,15 +1048,13 @@ class MySQLCursorRaw(MySQLCursor):
     _raw = True
 
     def fetchone(self):
-        row = self._fetch_row(raw=True)
-
-        if row:
-            return row
-        return None
+        self._check_executed()
+        return self._fetch_row(raw=True)
 
     def fetchall(self):
+        self._check_executed()
         if not self._have_unread_result():
-            raise errors.InterfaceError("No result set to fetch from.")
+            return []
         (rows, eof) = self._connection.get_rows(raw=True)
         if self._nextrow[0]:
             rows.insert(0, self._nextrow[0])
@@ -1071,14 +1085,11 @@ class MySQLCursorBufferedRaw(MySQLCursorBuffered):
             pass
 
     def fetchone(self):
-        row = self._fetch_row()
-        if row:
-            return row
-        return None
+        self._check_executed()
+        return self._fetch_row()
 
     def fetchall(self):
-        if self._rows is None:
-            raise errors.InterfaceError("No result set to fetch from.")
+        self._check_executed()
         return [r for r in self._rows[self._next_row:]]
 
     @property
@@ -1167,10 +1178,10 @@ class MySQLCursorPrepared(MySQLCursor):
             elif 'server_status' in result[2]:
                 self._handle_server_status(result[2]['server_status'])
 
-    def execute(self, operation, params=(), multi=False):  # multi is unused
+    def execute(self, operation, params=None, multi=False):  # multi is unused
         """Prepare and execute a MySQL Prepared Statement
 
-        This method will preare the given operation and execute it using
+        This method will prepare the given operation and execute it using
         the optionally given parameters.
 
         If the cursor instance already had a prepared statement, it is
@@ -1204,12 +1215,21 @@ class MySQLCursorPrepared(MySQLCursor):
 
         if self._prepared['parameters'] and not params:
             return
-        elif len(self._prepared['parameters']) != len(params):
-            raise errors.ProgrammingError(
-                errno=1210,
-                msg="Incorrect number of arguments " \
-                    "executing prepared statement")
+        elif params:
+            if not isinstance(params, (tuple, list)):
+                raise errors.ProgrammingError(
+                    errno=1210,
+                    msg=f"Incorrect type of argument: {type(params).__name__}({params})"
+                    ", it must be of type tuple or list the argument given to "
+                    "the prepared statement")
+            if len(self._prepared['parameters']) != len(params):
+                raise errors.ProgrammingError(
+                    errno=1210,
+                    msg="Incorrect number of arguments " \
+                        "executing prepared statement")
 
+        if params is None:
+            params = ()
         res = self._connection.cmd_stmt_execute(
             self._prepared['statement_id'],
             data=params,
@@ -1247,11 +1267,13 @@ class MySQLCursorPrepared(MySQLCursor):
 
         Returns a tuple or None.
         """
+        self._check_executed()
         if self._cursor_exists:
             self._connection.cmd_stmt_fetch(self._prepared['statement_id'])
         return self._fetch_row() or None
 
     def fetchmany(self, size=None):
+        self._check_executed()
         res = []
         cnt = (size or self.arraysize)
         while cnt > 0 and self._have_unread_result():
@@ -1262,8 +1284,7 @@ class MySQLCursorPrepared(MySQLCursor):
         return res
 
     def fetchall(self):
-        if not self._have_unread_result():
-            raise errors.InterfaceError("No result set to fetch from.")
+        self._check_executed()
         rows = []
         if self._nextrow[0]:
             rows.append(self._nextrow[0])
@@ -1305,6 +1326,7 @@ class MySQLCursorDict(MySQLCursor):
     def fetchone(self):
         """Returns next row of a query result set
         """
+        self._check_executed()
         row = self._fetch_row()
         if row:
             return self._row_to_python(row, self.description)
@@ -1313,8 +1335,10 @@ class MySQLCursorDict(MySQLCursor):
     def fetchall(self):
         """Returns all rows of a query result set
         """
+        self._check_executed()
         if not self._have_unread_result():
-            raise errors.InterfaceError(ERR_NO_RESULT_TO_FETCH)
+            return []
+
         (rows, eof) = self._connection.get_rows()
         if self._nextrow[0]:
             rows.insert(0, self._nextrow[0])
@@ -1359,6 +1383,7 @@ class MySQLCursorNamedTuple(MySQLCursor):
     def fetchone(self):
         """Returns next row of a query result set
         """
+        self._check_executed()
         row = self._fetch_row()
         if row:
             if hasattr(self._connection, 'converter'):
@@ -1369,8 +1394,10 @@ class MySQLCursorNamedTuple(MySQLCursor):
     def fetchall(self):
         """Returns all rows of a query result set
         """
+        self._check_executed()
         if not self._have_unread_result():
-            raise errors.InterfaceError(ERR_NO_RESULT_TO_FETCH)
+            return []
+
         (rows, eof) = self._connection.get_rows()
         if self._nextrow[0]:
             rows.insert(0, self._nextrow[0])
@@ -1392,6 +1419,7 @@ class MySQLCursorBufferedDict(MySQLCursorDict, MySQLCursorBuffered):
     def fetchone(self):
         """Returns next row of a query result set
         """
+        self._check_executed()
         row = self._fetch_row()
         if row:
             return self._row_to_python(row, self.description)
@@ -1400,7 +1428,7 @@ class MySQLCursorBufferedDict(MySQLCursorDict, MySQLCursorBuffered):
     def fetchall(self):
         """Returns all rows of a query result set
         """
-        if self._rows is None:
+        if self._executed is None or self._rows is None:
             raise errors.InterfaceError(ERR_NO_RESULT_TO_FETCH)
         res = []
         for row in self._rows[self._next_row:]:
@@ -1417,6 +1445,7 @@ class MySQLCursorBufferedNamedTuple(MySQLCursorNamedTuple, MySQLCursorBuffered):
     def fetchone(self):
         """Returns next row of a query result set
         """
+        self._check_executed()
         row = self._fetch_row()
         if row:
             return self._row_to_python(row, self.description)
@@ -1425,7 +1454,7 @@ class MySQLCursorBufferedNamedTuple(MySQLCursorNamedTuple, MySQLCursorBuffered):
     def fetchall(self):
         """Returns all rows of a query result set
         """
-        if self._rows is None:
+        if self._executed is None or self._rows is None:
             raise errors.InterfaceError(ERR_NO_RESULT_TO_FETCH)
         res = []
         for row in self._rows[self._next_row:]:
